library shelf.core;

import 'dart:async';
import 'dart:collection';
import 'dart:io';

import 'src/shared.dart';
export 'src/shared.dart';

class Shelf {
  final List<ShelfHandler> _handlers = new List<ShelfHandler>();

  Shelf add(ShelfHandler handler) {
    assert(handler != null);
    _handlers.add(handler);
    return this;
  }

  Future<ShelfResponse> handleRequest(HttpRequest request) =>
      _RequestFrame.handle(_getROHandlersCopy(), request);

  Future<ShelfServer> listen(int port) =>
      ShelfServer._listen(port, _getROHandlersCopy());

  List<ShelfHandler> _getROHandlersCopy() =>
      new UnmodifiableListView(new List.from(_handlers, growable: false));
}

class ShelfServer {
  final HttpServer _server;
  final int port;
  final List<ShelfHandler> _handlers;
  StreamSubscription<HttpRequest> _requestSub;

  ShelfServer._(this._server, this.port, this._handlers) {
    _requestSub = _server.listen(_handleRequest);
  }

  static Future<ShelfServer> _listen(int port, List<ShelfHandler> handlers) =>
      HttpServer.bind(InternetAddress.ANY_IP_V6, port)
        .then((HttpServer value) => new ShelfServer._(value, port, handlers));

  void _handleRequest(HttpRequest request) {
    HttpResponse response = request.response;
    _RequestFrame.handle(_handlers, request)
      .then((ShelfResponse value) {
        response.statusCode = value.code;
        response.write(value.body);
        return response.close();
      })
      .catchError((Object err, StackTrace stack) {
        response.statusCode = 500;
        response.writeln(err);
        response.writeln(stack);
        return response.close();
      });
  }
}

class _RequestFrame {
  final List<ShelfHandler> _handlers;
  final HttpRequest request;
  final Set<int> _ranHanders = new LinkedHashSet();

  _RequestFrame(this._handlers, this.request);

  static Future<ShelfResponse> handle(List<ShelfHandler> handlers,
      HttpRequest request) {

    var frame = new _RequestFrame(handlers, request);

    return frame._runHandler(request);
  }

  Future<ShelfResponse> _runHandler(HttpRequest request, [int index = -1]) {
    var handler = _getHandler(index);

    var next = () => _runHandler(request, index + 1);

    if(!_ranHanders.add(index)) {
      throw new StateError('Already ran the handler at index $index');
    }

    return new Future.sync(() => handler(request, next))
      .then((ShelfResponse response) {
        if(response != null) {
          return response;
        } else if(index == _handlers.length) {
          return null;
        } else {
          return next();
        }
      });
  }

  ShelfHandler _getHandler(int index) {
    if(index == -1) {
      return _defaultHandler;
    } else if(index == _handlers.length) {
      return _noopHandler;
    } else {
      return _handlers[index];
    }
  }
}

dynamic _noopHandler(HttpRequest request, Future<ShelfResponse> next()) => null;

dynamic _defaultHandler(HttpRequest request, Future<ShelfResponse> next()) =>
  next().then((ShelfResponse response) {
    if(response == null) {
      return new ShelfResponse.noop(request.uri);
    }
    return response;
  });

